import json
import subprocess
import re
import sys

from nittymcpick.cls.comment import Comment


class Linter():

    def __init__(self, _config, _appargs):
        self.__name = _config["linter"]["name"]
        self.__exec = _config["linter"]["path"]
        self.__args = _config["linter"]["args"]
        self.__appargs = _appargs
        self.__retregex = _config["linter"]["ret_regex"]
        self.__pattern = _config["matches"]["pattern"]
        self.__lineadjust = _config["linter"]["tweaks"]["line_count_adjust"]
        self.__singlefileexec = _config["linter"]["tweaks"]["single_file_exec"]

    def __eval(self, _in, _files):
        res = []
        for m in re.finditer(self.__retregex, _in, re.MULTILINE):
            _tpl = {
                "file": "Unknown file",
                "severity": "issue",
                "line": 1,
                "message": ""
            }
            for k, v in _tpl.items():
                if isinstance(v, int):
                    try:
                        _t = int(m.group(k).strip())
                    except (IndexError, ValueError):
                        _t = _tpl[k]
                    _tpl[k] = _t + self.__lineadjust
                else:
                    try:
                        _tpl[k] = m.group(k).strip()
                    except IndexError:
                        pass
            _f = [x for x in _files if x.file == _tpl["file"]]
            if _f:
                _f = _f[0]
                # if finding is not in change set, don't issue anything
                if (_tpl["line"] in _f.affectedlines and self.__appargs.onlynew) or (not self.__appargs.onlynew):
                    _msg = "{} found a potential {} - {}".format(
                        self.__name, _tpl["severity"], _tpl["message"])
                    res.append(Comment(
                        _msg, _tpl["line"], _f.relpath, _f.base_sha, _f.start_sha, _f.head_sha, self.__appargs.botname))
        return res

    def Run(self, _files):
        _matchfiles = [x for x in _files if re.match(self.__pattern, x.file)]
        if self.__singlefileexec:
            _loops = [[x] for x in _matchfiles]
        else:
            _loops = _matchfiles
        out = ""
        for l in _loops:
            try:
                out += subprocess.check_output([self.__exec] + self.__args + [x.file for x in l],
                                               universal_newlines=True, stderr=subprocess.STDOUT)
            except subprocess.CalledProcessError as e:
                out += e.stdout or ""
        return self.__eval(out, _matchfiles)

    @staticmethod
    def SetupLinter(_args):
        res = []
        __config = []
        try:
            with open(_args.config) as i:
                __config = json.load(i)
        except (json.JSONDecodeError, FileNotFoundError, PermissionError) as e:
            sys.stderr.write("Can't decode config: {}\n".format(e))
            sys.exit(-1)
        for c in __config:
            res.append(Linter(c, _args))
        return res



class Comment():

    def __init__(self, text, line, file, base_sha, start_sha, head_sha, username):
        self.__body = text
        self.__line = line
        self.__file = file
        self.__base_sha = base_sha
        self.__start_sha = start_sha
        self.__head_sha = head_sha
        self.__username = username

    def __repr__(self):
        return "{}: {}: {}".format(self.__file, self.__line, self.__body)

    @property
    def body(self):
        return self.__body

    @property
    def line(self):
        return self.__line

    def get_data(self):
        return {
            "body": self.__body,
            "position": {
                "base_sha": self.__base_sha,
                "start_sha": self.__start_sha,
                "head_sha": self.__head_sha,
                "position_type": "text",
                "new_line": self.__line,
                "new_path": self.__file
            },
            "author": {
                "username": self.__username
            }
        }

    def equals(self, disobj):
        _data = self.get_data()
        if _data["body"] == disobj["body"] and \
                _data["position"]["new_line"] == disobj["position"]["new_line"] and \
                _data["position"]["new_path"] == disobj["position"]["new_path"] and \
                _data["author"]["username"] == disobj["author"]["username"]:
            # print("Comment {} already exists".format(_data["body"]))
            return True
        return False



import os

from unidiff import PatchSet, UnidiffParseError


class DiffFile():

    def __init__(self, file, diff, base_sha, start_sha, head_sha, rootpath=""):
        self.__diff = diff
        self.__file = file
        self.__base_sha = base_sha
        self.__start_sha = start_sha
        self.__head_sha = head_sha
        self.__rootpath = rootpath
        self.__affectedlines = self.__get_modified_lines()

    def __get_modified_lines(self):
        res = []
        try:
            _patch = PatchSet(self.__diff)
        except UnidiffParseError:
            # On some version the diff header is missing
            # just add a fake one
            _patch = PatchSet("diff --git a/{} b/{}\n--- a/{}\n+++ b/{}\n{}".format(
                self.__file, self.__file, self.__file, self.__file, self.__diff))
        for _f in _patch.added_files + _patch.modified_files:
            for h in [x for x in _f]:
                res += h.target_lines()
        res = [x.target_line_no for x in res]
        return res

    def __getpath(self, relative=False):
        if not relative:
            if not self.__file.startswith("/"):
                return os.path.join(self.__rootpath, self.__file)
            return self.__file
        else:
            if not self.__file.startswith("/"):
                return self.__file
            return self.__file.replace(self.__rootpath, "", 1).lstrip("/")

    @property
    def file(self):
        return self.__getpath(relative=False)

    @property
    def relpath(self):
        return self.__getpath(relative=True)

    @property
    def diff(self):
        return self.__diff

    @property
    def base_sha(self):
        return self.__base_sha

    @property
    def affectedlines(self):
        return self.__affectedlines

    @property
    def start_sha(self):
        return self.__start_sha

    @property
    def head_sha(self):
        return self.__head_sha



import os
import subprocess
import sys
import shutil

from nittymcpick.cls.difffile import DiffFile


class Job():
    def __init__(self, event, gl, args, sgl, linter):
        self.__event = event
        self.__gl = gl
        self.__args = args
        self.__sgl = sgl
        self.__linter = linter

    def __repr__(self):
        return "Project: {}, MR: {}".format(self.__event.project_id, self.__event.object_attributes["iid"])

    def Run(self):
        if self.__event.object_attributes["work_in_progress"] and \
                not self.__args.nowip:
            print("Skip WIP for: {}".format(self))
            return
        if self.__event.object_attributes["state"] in ["merged", "closed"]:
            # Don't act on closed or merged MRs
            print("Skip closed for: {}".format(self))
            return

        _path = self.__checkout_source_branch(self.__args.tmpdir)
        _files = self.__get_commits_diffs(_path)
        if _path:
            _finding = self.__run_linter(_files)
            self.__lint_to_comments(_finding)
            self.__remove_clone(_path)

    def __get_commits_diffs(self, _path):
        mr = self.__sgl.projects.get(self.__event.project_id).mergerequests.get(
            self.__event.object_attributes["iid"])
        changes = mr.changes(all=True)
        return [DiffFile(x["new_path"], x["diff"], changes["diff_refs"]["base_sha"], changes["diff_refs"]["start_sha"], changes["diff_refs"]["head_sha"], _path)
                for x in mr.changes(all=True)["changes"]]

    def __checkout_source_branch(self, path):
        _path = f"{path}/{self.__event.project_id}/{self.__event.object_attributes['iid']}"
        good = False
        for method in [self.__event.object_attributes["source"]["ssh_url"], self.__event.object_attributes["source"]["http_url"]]:
            try:
                if os.path.exists(_path):
                    self.__remove_clone(_path)
                subprocess.check_call(
                    ["git", "clone", method, _path], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
                subprocess.check_call(["git", "-C", _path, "checkout", self.__event.object_attributes["last_commit"]["id"]],
                                      stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
                good = True
                break
            except subprocess.CalledProcessError:
                pass
        if not good:
            sys.stderr.write("Ohhhhhhh noooo - check out failed\n")
            return None
        return _path

    def __remove_clone(self, _path):
        shutil.rmtree(_path, ignore_errors=True)

    def __run_linter(self, _files):
        res = []
        for l in self.__linter:
            res += l.Run(_files)
        return res

    def __lint_to_comments(self, _input):
        mr = self.__sgl.projects.get(self.__event.project_id).mergerequests.get(
            self.__event.object_attributes["iid"])

        existing_comments = [mr.discussions.get(
            x.id) for x in mr.discussions.list(all=True)]
        __all_notes = []
        for e in existing_comments:
            __all_notes += e.attributes['notes']

        new_comments = []
        for f in _input:
            for e in existing_comments:
                __all_notes += e.attributes['notes']
            if not any([f.equals(x) for x in __all_notes]):
                if f.get_data() not in new_comments:
                    new_comments.append(f.get_data())

        resolved_comments = []
        for e in existing_comments:
            for n in e.attributes['notes']:
                if n["author"]["username"] != self.__args.botname:
                    continue
                if not any([x.equals(n) for x in _input]):
                    if (e, n["id"]) not in resolved_comments:
                        resolved_comments.append((e, n["id"]))
        print("Resolved {} issues for: {}".format(len(resolved_comments), self))
        print("New {} issues for: {}".format(len(new_comments), self))
        for f in new_comments:
            mr.discussions.create(f)

        for f in resolved_comments:
            try:
                f[0].notes.delete(f[1])
            except Exception as e:
                print(e)
        print("Done with {}".format(self))
